#include <QtGui>
#include <QtDebug>
#include "httpget.h"
#include "mapscene.h"

MapScene::MapScene(Model *model, QObject *parent) :
        QGraphicsScene(parent),
        myModel(model)
{
    connect(myModel, SIGNAL(mapChanged()), this, SLOT(redraw()));
    connect(&getter, SIGNAL(done(HttpGet*, bool)), this, SLOT(tileLoaded(HttpGet*, bool)));
}

QPixmap *MapScene::getPixmap(const QString &key) {
    foreach(PixmapEntry e, pixmaps) {
        if (e.key() == key)
            return e.pixmap();
    }
    return NULL;
}

void MapScene::redrawLayer(const QString &type, int z) {
    QString key;
    QString url;

    int x = myModel->x();
    int y = myModel->y();
    int w = myModel->width();
    int h = myModel->height();
    int offset = myModel->offset();
    for (int ix = 0; ix < w; ix++) {
        for (int iy = 0; iy < h; iy++) {
            int cx = x+ix*offset;
            int cy = y+iy*offset;
            key = QString("%1/%2_%3_%4").arg(type).arg(myModel->zoom())
                  .arg(cx, 0, 16).arg(cy, 0, 16);
            QPixmap *px = getPixmap(key);
            if (px != NULL) {
                addPixmap(px, ix*256, (h-1-iy)*256, z);
                continue;
            }
            requests.append(TileRequest(key, QPoint(ix*256, (h-1-iy)*256), z));
        }
    }
}

void MapScene::redraw() {
    QString key;
    QString url;

    qDebug()<<"redraw "<<myModel->type();
    foreach(QGraphicsItem *it, items()) {
        removeItem(it);
        delete it;
    }
    if (pixmaps.size() > 2000) {
        for (int i = 0; i < 1000; i++) {
            pixmaps.removeFirst();
        }
    }
    int w = myModel->width();
    int h = myModel->height();
    progressDlg = new QProgressDialog(tr("Download tiles"), tr("&Cancel"), 0, w*h);
    connect(progressDlg, SIGNAL(canceled()), this, SLOT(cancelRequests()));
    connect(progressDlg, SIGNAL(destroyed()), this, SLOT(cancelRequests()));
    progressDlg->setModal(true);
    progressDlg->show();
    redrawLayer(myModel->type(), 0);
    QList<Overlay> *ovls = myModel->overlays();
    foreach(Overlay ovl, *ovls) {
        redrawLayer(ovl.key(), ovl.z());
    }
    setSceneRect(0, 0, w*256, h*256);
    if (requests.size() == 0) {
        progressDlg->hide();
        delete progressDlg;
        progressDlg = NULL;
    }
    else {
        progressDlg->setMaximum(requests.size());
        getNextTile();
    }
}

void MapScene::addPixmap(QPixmap *pixmap, int ix, int iy, int z) {
    QGraphicsPixmapItem *it = new QGraphicsPixmapItem(*pixmap);
    it->setOffset(ix, iy);
    it->setZValue(z);
    addItem(it);
}

void MapScene::getNextTile() {
    while(requests.size() > 0) {
        TileRequest req = requests[0];
        QString url = QString("%1/%2").arg(URL_MAPY_CZ).arg(req.key());
        qDebug()<<url;
        if (getter.getFile(QUrl(url), req)) {
            return;
        }
        else {
            requests.removeAt(0);
        }
    }
    progressDlg->hide();
    delete progressDlg;
    progressDlg = NULL;
}

void MapScene::tileLoaded(HttpGet *getter, bool error) {
    TileRequest req = getter->request();
    qDebug()<<"tileLoaded "<<req.pos();
    if (requests.contains(req)) {
        requests.removeOne(req);
        if (!error) {
            QPoint p = req.pos();
            QPixmap *pixmap = new QPixmap();
            bool res = pixmap->loadFromData(getter->getData());
            getter->close();
            if (res) {
                pixmaps.append(PixmapEntry(req.key(), pixmap));
                addPixmap(pixmap, p.x(), p.y(), req.z());
            }
            else {
                delete pixmap;
            }
        }
    }
    progressDlg->setValue(progressDlg->maximum()-requests.size());
    getNextTile();
    //progressDlg->setValue(getters.size());
}

void MapScene::cancelRequests() {
    if (requests.size() == 0)
        return;
    qDebug()<<"cancelRequests";
    getter.abort();
    requests.clear();
}
